module Main where

import Concurrent
import Exception
import IO hiding (catch)
import Monad (unless)
import Optimize (optimizeDocument)
import Parsec
import Parser
import Prelude hiding (catch)
import SML
import System
import Time

outputTime = 5

main :: IO ()
main = do args <- getArgs
          let timeout = if length args == 1 then 
                            read (head args) 
                        else 
                            180::Int

          contents <- getContents
          parseResult <- newEmptyMVar
          let parserAction = storeResult parseResult (timeAction (parseInput contents))
          parseCompleted <- forkTimedIO (timeout - outputTime) parserAction
          hPutStrLn stderr ("parseCompleted: " ++ show parseCompleted)
          case parseCompleted of
              False -> exitFailure --putStr contents
              True  -> do (parseTime, doc) <- takeMVar parseResult
                          hPutStrLn stderr (show parseTime)
                          resultDoc <- newMVar doc
                          forkTimedIO (timeout - parseTime - outputTime) (optimize resultDoc)
                          finalDoc <- takeMVar resultDoc
                          putStr (show finalDoc)

--  parseInput :: String -> IO Document
--  parseInput contents = case parse parseSML "" contents of
--                            Left err -> error (show err)
--                            Right doc -> return $! (length (show doc) `seq` doc)

parseInput :: String -> IO Document
parseInput contents = docSize doc `seq` return doc
    where doc = mainParser contents
 
ignoreException :: IO () -> IO ()
ignoreException action
     = action `catch` \ex -> do hPutStrLn stderr ("EXCEPTION: " ++ show ex); return ()

storeResult :: MVar a -> IO a -> IO ()
storeResult result action = do v <- action; putMVar result v

forkTimedIO :: Int -> IO () -> IO Bool
forkTimedIO timeout thread = do
     mvar <- newEmptyMVar
     threadId <- forkIO (ignoreException thread `finally` putMVar mvar ())
     currentTime <- getClockTime
     let stopTime = addToClockTime (TimeDiff { tdYear = 0, tdMonth = 0, tdDay = 0, tdHour = 0, tdMin = 0, tdSec = timeout, tdPicosec = 0 })
                                   currentTime
     completed <- waitMVar stopTime mvar
     unless completed (killThread threadId)
     takeMVar mvar
     return completed

waitMVar :: ClockTime -> MVar () -> IO Bool
waitMVar stopTime mvar = do
    currentTime <- getClockTime
    if currentTime >= stopTime then return False
     else do hPutStrLn stderr ("Timeout: " ++ show (diffClockTimes stopTime currentTime))
             result <- tryTakeMVar mvar
             maybe (do threadDelay 1000000
                       waitMVar stopTime mvar)
                   (\v -> do putMVar mvar v
                             return True)
                   result

-- Times action and returns the number of elapsed seconds.
timeAction :: IO a -> IO (Int, a)
timeAction action = do
    start <- getClockTime
    result <- action
    stop <- result `seq` getClockTime
    return (timeDiffToSeconds (diffClockTimes stop start), result)
                  

timeDiffToSeconds :: TimeDiff -> Int
timeDiffToSeconds td =
    tdHour td * 3600 + tdMin td * 60 + tdSec td

optimize :: MVar Document -> IO ()
optimize doc = do optimizeDocument doc
