module SML (
    Colors(..),
    Tags(..),
    Context, b, em, i, s, tt, u, size, color,
    Document(..),
    TagTransition(..),
    TagOp(..),
    docDepth,
    rootContext, 
    applyTag, 
    makeText, makeTag, getText, isText, mergeText,
    tagSize, docSize,
    isSpaceContextEq, isEffective, isWhiteSpace,
    updateChildren, updateTag, childContext,
    hideEffectOf, allTags, findMultiTagTransition, applyTrans, sizeTagOp,
    sizeTagTransition,ordTagTransition,
    minTagTransition, meaning, updateContext, updateContexts,
    findOpenMultiTagTransition
) where

--  import Char (ord, chr)
--  import IArray
import List
--  import UnitTest
--  import Word

--  boldItalContext = rootContext { b = True, i = True }
--  redContext = rootContext { color = Red }
--  greenContext = rootContext { color = Green }
--  blueContext = rootContext { color = Blue }
--  emContext = rootContext { em = True }
--  uContext = rootContext { u = 1 }
--  uEmContext = applyTag uContext EM

--  docs1 = [(Text redContext "red", redContext), (Text greenContext "green", greenContext), (Text blueContext "blue", blueContext), (Text redContext "red", redContext)]
--  docs2 = [(Tag emContext EM [Text rootContext "Q"], emContext)]
--  doc1 = Tag rootContext Root [Tag rootContext U [Tag uContext EM [Text uEmContext "f"]]]
--  doc3 = Tag rootContext EM [Tag emContext EM [Text rootContext "Q"]]

data Colors = Red | Green | Blue | Cyan | Magenta | Yellow | K | White | UnknownColor
            deriving (Eq)

instance Show Colors where
    showsPrec _ Red x     = "r" ++ x
    showsPrec _ Green x   = "g" ++ x
    showsPrec _ Blue x    = "b" ++ x
    showsPrec _ Cyan x    = "c" ++ x
    showsPrec _ Magenta x = "m" ++ x
    showsPrec _ Yellow x  = "y" ++ x
    showsPrec _ K x       = "k" ++ x
    showsPrec _ White x   = "w" ++ x
    showsPrec _ UnknownColor x = "unknown color" ++ x

data Tags = Root | B | EM | I | PL | S | TT | U | Size Int | Color Colors
          deriving (Eq)

instance Show Tags where
    showsPrec _ Root x      = "root tag" ++ x
    showsPrec _ B x         = "B" ++ x
    showsPrec _ EM x        = "EM" ++ x
    showsPrec _ I x         = "I" ++ x
    showsPrec _ PL x        = "PL" ++ x
    showsPrec _ S x         = "S" ++ x
    showsPrec _ TT x        = "TT" ++ x
    showsPrec _ U x         = "U" ++ x
    showsPrec _ (Size s) x  = show s ++ x
    showsPrec _ (Color c) x = show c ++ x

data Context = Context { 
    b     :: !Bool,
    em    :: !Bool,
    i     :: !Bool,
    s     :: !Bool,
    tt    :: !Bool,
    u     :: !Int,
    size  :: !Int,
    color :: !Colors 
} deriving (Eq, Show)


data Document = Tag { context  :: !Context,
                      tag      :: !Tags,
                      children :: ![Document] }
              | Text { context :: !Context,
                       text    :: !String }
              deriving (Eq)

instance Show Document where
    showsPrec n t x = docShowsPrec n t x

docShowsPrec n (Tag _ Root ds) x 
        = foldr (docShowsPrec n) x ds
docShowsPrec n (Tag _ tag ds) x  
        = "<" ++ show tag ++ ">" ++ 
          foldr (docShowsPrec n) ("</" ++ show tag ++ ">" ++ x) ds
docShowsPrec n (Text _ s) x      = s ++ x

rootContext :: Context
rootContext = Context { b = False, em = False, i = False, s = False, tt = False, u = 0, size = -1, color = UnknownColor }

childContext :: Document -> Context
childContext (Tag ctx tag _) = applyTag ctx tag
childContext _ = error "no child context"

tagSize :: Tags -> Int
tagSize B = 7
tagSize EM = 9
tagSize I = 7
tagSize PL = 9
tagSize S = 7
tagSize TT = 9
tagSize U = 7
tagSize (Size _) = 7
tagSize (Color _) = 7
tagSize Root = 0

allTags = [ B, EM, I, PL, S, TT, U, 
            Size 0, Size 1, Size 2, Size 3, Size 4,
            Size 5, Size 6, Size 7, Size 8, Size 9,
            Color Red, Color Green, Color Blue,
            Color Cyan, Color Magenta, Color Yellow,
            Color K, Color White ]

docSize :: Document -> Int
docSize (Tag _ tag children) = tagSize tag + sum (map docSize children)
docSize (Text _ s) = length s

docDepth :: Document -> Int
docDepth (Text _ _)         = 0
docDepth (Tag _ _ children) = 1 + maximum (0 : map docDepth children)

maxChildLength :: Document -> Int
maxChildLength (Text _ _)         = 0
maxChildLength (Tag _ _ children) 
    = maximum (length children : map maxChildLength children)

-- TODO: fix EM
applyTag :: Context -> Tags -> Context
applyTag context EM 
    | s context = context
    | otherwise = context { em = not (em context) }
applyTag context B  = context { b = True }
applyTag context I  = context { i = True }
applyTag context PL = context { u = 0, b = False, em = False, i = False, s = False, tt = False }
applyTag context S  = context { s = True, em = False }
applyTag context TT = context { tt = True }
applyTag context U  = context { u = min 3 (1 + u context) }
applyTag context (Size s) = context { size = s }
applyTag context (Color c) = context { color = c }
applyTag context Root = context `seq` context

makeText :: Context -> String -> Document
makeText context string = Text context (compressSpaces context string)

mergeText :: Document -> Document -> Document
mergeText (Text c1 s1) (Text c2 s2) = makeText c1 (s1++s2)
mergeText _ _ = error "bad merge text"

isText :: Document -> Bool
isText (Text _ _) = True
isText _ = False

getText :: Document -> String
getText (Text _ s) = s
getText _ = error "no match"

isSpaceContextEq :: Context -> Context -> Bool
isSpaceContextEq c1 c2
    | tt c1 || tt c2         = False
    | size c1 /= size c2     = False
    | u c1 /= u c2           = False
    | u c1 == 0 && u c2 == 0 = True
    | color c1 /= color c2   = False
    | otherwise              = True

makeSpaceContext :: Context -> Context
makeSpaceContext ctx
    | u ctx == 0 = ctx { s = False, i = False, em = False, b = False, color = White }
    | u ctx /= 0 = ctx { s = False, i = False, em = False, b = False }

isEffective :: Context -> Tags -> Bool
isEffective c Root = error "is effective on root"
isEffective c B  = not (b c)
isEffective c EM = not (s c)
isEffective c I  = not (i c)
isEffective c PL = b c || em c || i c || s c || tt c || u c /= 0
isEffective c S  = not (s c)
isEffective c TT = not (tt c)
isEffective c U  = u c < 3
isEffective c (Size s)    = size c /= s
isEffective c (Color col) = color c /= col

countPLFlags :: Context -> Int
countPLFlags c = length (filter (==False)
                                [ b c, em c, i c, s c, tt c, u c /= 0 ])

contextCloser :: Context -> Tags -> Context -> Bool
contextCloser from U to   = u from < u to
contextCloser from EM to  = em from /= em to
contextCloser from PL to  = countPLFlags from < countPLFlags to
contextCloser from tag to = not (isEffective to tag)

makeTag :: Context -> Tags -> [Document] -> [Document]
makeTag context tag children = context `seq` tag `seq` children `seq`
    case children of
    [] -> []
    _  -> case isEffective context tag of
          True -> [Tag context tag children]
          False -> children



updateChildren :: Document -> [Document] -> Document
updateChildren (Tag context tag _) children
    = children `seq` Tag context tag children
updateChildren _ _ = error "not a tag"

updateTag :: Document -> Tags -> Document
updateTag (Text context s) tag = Text (applyTag context tag) s
updateTag _ _ = error "updateTag not applied to text"

updateContext :: Context -> Document -> Document
updateContext ctx doc = doc { context = ctx }

updateContexts ctx docs = map (updateContext ctx) docs

isWhiteSpace :: Char -> Bool
isWhiteSpace '\x20' = True
isWhiteSpace '\x0d' = True
isWhiteSpace '\x0a' = True
isWhiteSpace '\x09' = True
isWhiteSpace _     = False

compressSpaces :: Context -> String -> String
compressSpaces context [] = context `seq` []
compressSpaces context s@(x:xs) 
    | tt context     = s
    | isWhiteSpace x = x : compressSpacesWS xs
    | otherwise      = x : compressSpacesNoWS xs

compressSpacesWS [] = []
compressSpacesWS (x:xs)
    | isWhiteSpace x = compressSpacesWS xs
    | otherwise      = x : compressSpacesNoWS xs

compressSpacesNoWS [] = []
compressSpacesNoWS (x:xs)
    | isWhiteSpace x = x : compressSpacesWS xs
    | otherwise      = x : compressSpacesNoWS xs

meaning :: Context -> Context -> Document -> Bool -> [(Char, Context)]
meaning prev cur v@(Text _ []) ws = []
meaning prev cur v@(Text _ (s:ss)) ws
   | tt cur && isWhiteSpace s
       = (s, makeSpaceContext cur) : meaning (makeSpaceContext cur) cur (Text cur ss) False
   | isWhiteSpace s && ws && isSpaceContextEq prev cur
       = meaning prev cur (Text cur ss) True
   | isWhiteSpace s
       = (' ', makeSpaceContext cur) : meaning (makeSpaceContext cur) cur (Text cur ss) True
   | otherwise = (s, cur) : meaning cur cur (Text cur ss) False
meaning prev cur v@(Tag _ t []) ws = []
meaning prev cur v@(Tag _ t (c:cc)) ws
   = let gen = meaning prev (applyTag cur t) c ws
         lastws = if null gen then ws else isWhiteSpace (fst (last gen))
         lastctx = if null gen then prev else snd (last gen)
     in  gen ++ meaning lastctx cur (Tag cur t cc) lastws


-- True if t1 hides the effects of t2.
hideEffectOf :: Tags -> Tags -> Bool
hideEffectOf (Size _) (Size _) = True
hideEffectOf (Color _) (Color _) = True
hideEffectOf PL U = True
hideEffectOf PL S = True
hideEffectOf PL EM = True
hideEffectOf PL I = True
hideEffectOf PL B = True
hideEffectOf PL TT = True
hideEffectOf S EM = True
hideEffectOf _ _ = False

--  findTagTransitions :: [Tags] -> Context -> Context -> [([Tags], [Tags])]
--  findTagTransitions open from to
--       = findTagTransitionsOpenB open from to
--         ++ findTagTransitionsCloseB open from to

--  findTagTransitionsOpenB open from to
--      | b from     = []
--      | not (b to) = []
--      | otherwise  = [([], [B])]

--  findTagTransitionsCloseB (B:rest) from to
--      | not (b from) = error "b not open!"
--      | b to         = []
--      | applyTag from PL /= to = [([B], [])]
--      | otherwise    = [([B], []), ([], [PL])]
--  findTagTransitionsCloseB open from to
--      | not (b from) = []
--      | b to         = []
--      | applyTag from PL == to = [([], [PL])]
--      | otherwise    = []

findOpenTagTransition :: Context -> Context -> [[Tags]]
findOpenTagTransition from to
    | from == to = []
    | otherwise  = [ [x] | x <- allTags, applyTag from x == to ]

--  findSingleTagTransition [] from to
--      = findOpenTagTransition from to
--  findSingleTagTransition ((tag, old):open) from to
--      | old == to = [([tag], [])] ++ openPossibilities
--      | otherwise = openPossibilities
--      where openPossibilities = findOpenTagTransition from to

data TagOp         = Push Tags | Pop Tags | Doc Document deriving (Eq, Show)
type TagTransition = [TagOp]

sizeTagTransition :: TagTransition -> Int
sizeTagTransition ts = sum (map sizeTagOp ts)

minTagTransition :: TagTransition -> TagTransition -> TagTransition
minTagTransition t1 t2
    | sizeTagTransition t1 <= sizeTagTransition t2 = t1
    | sizeTagTransition t1 > sizeTagTransition t2 =  t2

ordTagTransition :: TagTransition -> TagTransition -> Ordering
ordTagTransition t1 t2
    | sizeTagTransition t1 < sizeTagTransition t2 = LT
    | sizeTagTransition t1 == sizeTagTransition t2 = EQ
    | sizeTagTransition t1 > sizeTagTransition t2 = GT

sizeTagOp (Push t) = (tagSize t - 1)
sizeTagOp (Pop t)  = 1
sizeTagOp (Doc d)  = 0

applyTrans :: Context -> [(Tags, Context)] -> [TagOp] -> [(Tags, Context)]
applyTrans ctx (c:cs) (Pop _:rest) 
    = applyTrans (snd c) cs rest
applyTrans ctx orig (Push tag:rest) 
    = applyTrans (applyTag ctx tag) ((tag, ctx):orig) rest
applyTrans ctx s [] = s

findMultiTagTransition :: [(Tags, Context)] -> Context -> Context -> [TagTransition]
findMultiTagTransition open from to
    = [ (map (Pop . fst) pops) ++ (map Push opened)
      | pops   <- reverse (inits open),
        opened <- findOpenMultiTagTransition 
                      [] 
                      (if null pops then from else (snd (last pops)))
                      to ]
                      

--  findMultiTagTransition :: Int -> [(Tags, Context)] -> Context -> Context -> [([Tags], [Tags])]
--  findMultiTagTransition 0 _ _ _ = []
--  findMultiTagTransition n open from to
--      = [  | t <- allTags,
--                  rests <- findMultiTagTransition (n - 1) 
--                                                  open (applyTag from t) to ]

findOpenMultiTagTransition :: [Tags] -> Context -> Context -> [[Tags]]
findOpenMultiTagTransition gen from to
    | from == to = [[]]
    | otherwise  
        = let single = findOpenTagTransition from to
          in  case single of
              [] -> [ t:opened
                    | t <- allTags, 
                      t `notElem` gen,
                      isEffective from t,
                      contextCloser from t to,
                      opened <- findOpenMultiTagTransition (t:gen)
                                                           (applyTag from t) 
                                                           to ]
              _  -> single

--  findMultiTagTransition open from to
--      | from == to = []
--      | otherwise  = [ p 
--                     | t <- allTags,
--                       p <- findSingleTagTransition ((t, from):open) 
--                                                    (applyTag from t)
--                                                    to ]

-- Tests

--  testSML = testSuite "SML" [ testApplyTag, testIsSpaceContextEq, testCompressSpaces ]

--  testApplyTag = testCase "applyTag" (do
--      let context = applyTag rootContext U
--      checkEq "U != 1" 1 (u context)
--      let context' = applyTag (applyTag context U) U
--      checkEq "U != 3" 3 (u context')
--      let context'' = applyTag context' U
--      checkEq "U != 3" 3 (u context''))

--  testIsSpaceContextEq = testCase "isSpaceContextEq" (do
--      let c1 = Context { b = True, em = False, i = False, s = False, tt = False, u = 0, size = -1, color = UnknownColor }
--      let c2 = Context { b = True, em = False, i = True, s = False, tt = False, u = 0, size = -1, color = UnknownColor }
--      check "c1 not space eq with c2" (isSpaceContextEq c1 c2)
--      let c3 = Context { b = True, em = False, i = True, s = True, tt = False, u = 0, size = -1, color = UnknownColor }
--      check "c1 not space eq with c3" (isSpaceContextEq c1 c3)
--      let c4 = Context { b = True, em = False, i = True, s = True, tt = True, u = 0, size = -1, color = UnknownColor }
--      check "c1 space eq with c4" (not $ isSpaceContextEq c1 c4)
--      check "c4 space eq with c4" (not $ isSpaceContextEq c4 c4)
--      )

--  testCompressSpaces = testCase "compressSpaces" (do
--      let c1 = rootContext
--      let c2 = c1 { tt = True }
--      let s1 = " hello    there   f   "
--      let s2 = " hello there f "
--      checkEq "whitespace not compressed" s2 (compressSpaces c1 s1)
--      checkEq "whitespace compressed" s1 (compressSpaces c2 s1))
