
import java.net.*;
import java.io.*;

public class CompareDocs {

    public static final String OK_STRING = "Documents are equivalent";
    public static final String DIFF_STRING = "Documents are not equivalent";
    public static final String SCRIPT = "http://cristal.inria.fr/bin/compare";

    String file1;
    String file2;

    String doc1;
    String doc2;
    
    public static void main(String[] args) throws Exception {

        if(args.length < 2) {
            System.err.println("Provide 2 filenames");
            System.exit(1);
        }
        
        CompareDocs cmp = new CompareDocs(args[0], args[1]);
        TestResult result = cmp.doTest(args[0]);

        System.out.println(result);
        if(result.equiv) {
            System.exit(0);
        } else {
            System.exit(1);
        }
    }

    public CompareDocs(String f1, String f2) throws Exception {
        // read files into doc1 & doc2
        doc1 = readFile(f1);
        doc2 = readFile(f2);
    }

    protected String readFile(String f) throws Exception {
        BufferedReader br1 = new BufferedReader(new FileReader(f));
        StringBuffer sb1 = new StringBuffer();
        String line;
        int c;
        
        while((c = br1.read()) != -1) {
            sb1.append( (char) c);
        }
        return sb1.toString();
    }
    
    public TestResult doTest(String input) throws Exception {
        URL url = new URL(SCRIPT);
        URLConnection con = url.openConnection();
        con.setDoOutput(true);
        PrintStream out = new PrintStream(con.getOutputStream());
        out.print("documentone=" + URLEncoder.encode(doc1) + "&documenttwo=" + URLEncoder.encode(doc2));
        out.close();

        String line = null;
        String result = null;
        boolean equiv = false;

        DataInputStream in =
            new DataInputStream(con.getInputStream());

        while((line = in.readLine()) != null) {
            if(line.startsWith(OK_STRING))
                equiv = true;
            else if(line.indexOf(DIFF_STRING) >= 0) {
                result = line;
            }
        }

        in.close();

        return new TestResult(input, equiv, result, doc1.length(), doc2.length());
    }

    private class TestResult {

        public String in;
        
        public boolean equiv = false;
        public String diff = "";
        public int l1;
        public int l2;

        public TestResult(String in, boolean b, String s) {
            this.in = in;
            equiv = b;
            diff = s;
        }

        public TestResult(String in, boolean b, String s, int i1, int i2) {
            this(in, b, s);
            l1 = i1;
            l2 = i2;
        }

        public String toString() {
            if(equiv) {
                return new String(in + ": OK (size diff: " + l1 + " vs. " + l2 + ", " + (l2 == l1 ? 100 : (l2 * 100 / l1)) + "% )");
            } else {
                return new String(in + ": DIFFERENT - " + diff);
            }
        }
    }
}
