
import java.net.*;
import java.io.*;
import java.util.*;

public class GenDocs {

    int MAX_DOCUMENTS = 15;
    int MAX_DEPTH     = 15;
    
    Random rand;

    int items;
    int depth;
    
    static String[] openTags = { "<B>", "<EM>", "<I>", "<PL>", "<S>", "<TT>", "<U>", "<0>",
                                 "<1>", "<2>", "<3>", "<4>", "<5>", "<6>", "<7>", "<8>", "<9>",
                                 "<r>", "<g>", "<b>", "<c>", "<m>", "<y>", "<k>", "<w>" };

    static String[] closeTags ={ "</B>", "</EM>", "</I>", "</PL>", "</S>", "</TT>", "</U>", "</0>",
                                 "</1>", "</2>", "</3>", "</4>", "</5>", "</6>", "</7>", "</8>", "</9>",
                                 "</r>", "</g>", "</b>", "</c>", "</m>", "</y>", "</k>", "</w>" };
    
    public static void main(String[] args) throws Exception {
        GenDocs d;
        
        if(args.length == 0) {
            d = new GenDocs();
        } else {
            int items = Integer.valueOf(args[0]).intValue();
            int depth = Integer.valueOf(args[1]).intValue();
            d = new GenDocs(items, depth);
        }
        
        d.generate();
    }

    public GenDocs() {
        rand = new Random();

        this.items = rand.nextInt(MAX_DOCUMENTS);
        this.depth = rand.nextInt(MAX_DEPTH);
    }

    public GenDocs(int items, int depth) {
        this();
        
        this.items = items;
        this.depth = depth;
    }
    
    public void generate() throws Exception {
        BufferedWriter bw = new BufferedWriter( new PrintWriter( System.out ) );
        bw.write("nl");
        generateDocuments(bw, depth, items);
        bw.write("nl");
        bw.close();
    }

    public void generateDocuments(Writer w, int depth, int items) throws Exception {
        for(int i=0; i<items; i++)
            generateDocument(w, depth, items - i);
    }

    public void generateDocument(Writer w, int depth, int items) throws Exception {
        if(depth <= 1) {
            generateLeaf(w);
        } else {
            int nextItem = rand.nextInt(100);
            
            if(nextItem <= 40) {
                generateTag(w, depth, items - 1);
            } else if(nextItem >= 60) {
                generateDocuments(w, depth, rand.nextInt(items));
            } else
                generateLeaf(w);
        }
    }

    public void generateTag(Writer w, int depth, int items) throws Exception {
        int tagIdx = rand.nextInt(openTags.length);

        writeString(w, openTags[tagIdx]);

        generateDocuments(w, depth - 1, items);

        writeString(w, closeTags[tagIdx]);
    }

    public void generateLeaf(Writer w) throws Exception {
        if(rand.nextInt(100) > 75)
            generateText(w);
        else
            generateWS(w);
    }
    
    public void generateText(Writer w) throws Exception {
        StringBuffer sb = new StringBuffer();
        int len = rand.nextInt(20);
        for(int i=0; i<len; i++) {
            char c = (char) (33 + rand.nextInt(127 - 33));
            if(c != '<' && c != '>' && c != '%')
                sb.append( c );
        }
        writeString(w, sb.toString());
    }

    public void generateWS(Writer w) throws Exception {
        StringBuffer sb = new StringBuffer();
        int len = rand.nextInt(20);

        for(int i=0; i<len; i++) {
            int r = rand.nextInt(100);
            char c = ' ';
            
            if(r == 14)
                c = '\t';

            if(r > 73)
                c = '\n';

            if(r == 72)
                c = '\r';
            
            sb.append(c);
        }
        writeString(w, sb.toString());
    }

    protected void writeString(Writer w, String s) throws Exception {
        w.write(s, 0, s.length());
    }
}
